from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from publsp.api.session import session_manager
from publsp.api.routes import session, ads, orders, channels

app = FastAPI(
    title="publsp API",
    description="API for purchasing liquidity from an LSP "
    "coordinated over Nostr"
)

# Add CORS middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=["http://localhost:3000", "http://localhost:5173", "http://127.0.0.1:3000", "http://127.0.0.1:5173"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Include routers
app.include_router(session.router)
app.include_router(ads.router)
app.include_router(orders.router)
app.include_router(channels.router)


# Startup and shutdown events
@app.on_event("startup")
async def startup_event():
    # Start the session maintenance task
    await session_manager.start_maintenance()


@app.on_event("shutdown")
async def shutdown_event():
    # Properly clean up all sessions
    await session_manager.shutdown()



# ripped from
# https://github.com/rustyrussell/lightning-payencode/blob/master/lnaddr.py and
# adapted for the purpose of decoding BOLT11 invoices on the
# customer/client-side so they don't need a LN node just to browse and purchase
# channels

from publsp.ln.bech32 import bech32_decode, CHARSET
from binascii import hexlify
from decimal import Decimal

import bitstring
import re
import secp256k1
import time


def unshorten_amount(amount):
    """ Given a shortened amount, convert it into a decimal
    # BOLT #11:
    # The following `multiplier` letters are defined:
    #
    #* `m` (milli): multiply by 0.001
    #* `u` (micro): multiply by 0.000001
    #* `n` (nano): multiply by 0.000000001
    #* `p` (pico): multiply by 0.000000000001
    """
    units = {
        'p': 10**12,
        'n': 10**9,
        'u': 10**6,
        'm': 10**3,
    }
    unit = str(amount)[-1]
    # BOLT #11:
    # A reader SHOULD fail if `amount` contains a non-digit, or is followed by
    # anything except a `multiplier` in the table above.
    if not re.fullmatch(r"\d+[pnum]?", str(amount)):
        raise ValueError("Invalid amount '{}'".format(amount))

    if unit in units.keys():
        return Decimal(amount[:-1]) / units[unit]
    else:
        return Decimal(amount)


# Bech32 spits out array of 5-bit values.  Shim here.
def u5_to_bitarray(arr):
    ret = bitstring.BitArray()
    for a in arr:
        ret += bitstring.pack("uint:5", a)
    return ret


def bitarray_to_u5(barr):
    assert barr.len % 5 == 0
    ret = []
    s = bitstring.ConstBitStream(barr)
    while s.pos != s.len:
        ret.append(s.read(5).uint)
    return ret


# Tagged field containing BitArray
def tagged(char, l):
    # Tagged fields need to be zero-padded to 5 bits.
    while l.len % 5 != 0:
        l.append('0b0')
    return bitstring.pack("uint:5, uint:5, uint:5",
                          CHARSET.find(char),
                          (l.len / 5) / 32, (l.len / 5) % 32) + l


# Discard trailing bits, convert to bytes.
def trim_to_bytes(barr):
    # Adds a byte if necessary.
    b = barr.tobytes()
    if barr.len % 8 != 0:
        return b[:-1]
    return b


# Try to pull out tagged data: returns tag, tagged data and remainder.
def pull_tagged(stream):
    tag = stream.read(5).uint
    length = stream.read(5).uint * 32 + stream.read(5).uint
    return (CHARSET[tag], stream.read(length * 5), stream)


def lndecode(a):
    hrp, data = bech32_decode(a)
    if not hrp:
        raise ValueError("Bad bech32 checksum")

    # BOLT #11:
    #
    # A reader MUST fail if it does not understand the `prefix`.
    if not hrp.startswith('ln'):
        raise ValueError("Does not start with ln")

    data = u5_to_bitarray(data)

    # Final signature 65 bytes, split it off.
    if len(data) < 65*8:
        raise ValueError("Too short to contain signature")
    sigdecoded = data[-65*8:].tobytes()
    data = bitstring.ConstBitStream(data[:-65*8])

    addr = LnAddr()
    addr.pubkey = None

    m = re.search(r"[^\d]+", hrp[2:])
    if m:
        addr.currency = m.group(0)
        amountstr = hrp[2+m.end():]
        # BOLT #11:
        #
        # A reader SHOULD indicate if amount is unspecified, otherwise it MUST
        # multiply `amount` by the `multiplier` value (if any) to derive the
        # amount required for payment.
        if amountstr != '':
            addr.amount = unshorten_amount(amountstr)

    addr.date = data.read(35).uint

    while data.pos != data.len:
        tag, tagdata, data = pull_tagged(data)
        # BOLT #11:
        #
        # * `r` (3): `data_length` variable.  One or more entries
        # containing extra routing information for a private route;
        # there may be more than one `r` field, too.
        #    * `pubkey` (264 bits)
        #    * `short_channel_id` (64 bits)
        #    * `feebase` (32 bits, big-endian)
        #    * `feerate` (32 bits, big-endian)
        #    * `cltv_expiry_delta` (16 bits, big-endian)
        route = []
        s = bitstring.ConstBitStream(tagdata)
        while s.pos + 264 + 64 + 32 + 32 + 16 < s.len:
            route.append((s.read(264).tobytes(),
                          s.read(64).tobytes(),
                          s.read(32).intbe,
                          s.read(32).intbe,
                          s.read(16).intbe))
        addr.tags.append(('r', route))

    # BOLT #11:
    #
    # A reader MUST check that the `signature` is valid (see the `n` tagged
    # field specified below).
    if addr.pubkey:  # Specified by `n`
        # BOLT #11:
        #
        # A reader MUST use the `n` field to validate the signature instead of
        # performing signature recovery if a valid `n` field is provided.
        addr.signature = addr.pubkey.ecdsa_deserialize_compact(sigdecoded[0:64])
        if not addr.pubkey.ecdsa_verify(bytearray([ord(c) for c in hrp]) + data.tobytes(), addr.signature):
            raise ValueError('Invalid signature')
    else:  # Recover pubkey from signature.
        addr.pubkey = secp256k1.PublicKey()
        addr.signature = addr.pubkey.ecdsa_recoverable_deserialize(
            sigdecoded[0:64], sigdecoded[64])
        addr.pubkey.public_key = addr.pubkey.ecdsa_recover(
            bytearray([ord(c) for c in hrp]) + data.tobytes(), addr.signature)

    return addr


class LnAddr(object):
    def __init__(
            self,
            paymenthash=None,
            amount=None,
            currency='bc',
            tags=None,
            date=None):
        self.date = int(time.time()) if not date else int(date)
        self.tags = [] if not tags else tags
        self.unknown_tags = []
        self.paymenthash = paymenthash
        self.signature = None
        self.pubkey = None
        self.currency = currency
        self.amount = amount

    def __str__(self):
        return "LnAddr[{}, amount={}{} tags=[{}]]".format(
            hexlify(self.pubkey.serialize()).decode('utf-8'),
            self.amount, self.currency,
            ", ".join([k + '=' + str(v) for k, v in self.tags])
        )



from fastapi import APIRouter, Depends
from pydantic import BaseModel

from publsp.api.session import UserSession
from publsp.api.utils import get_user_session


class SessionInfo(BaseModel):
    session_id: str
    nostr_pubkey: str
    created_at: str
    last_accessed: str


router = APIRouter(prefix="/session", tags=["Session"])


@router.get("/", response_model=SessionInfo)
async def get_session(session: UserSession = Depends(get_user_session)):
    """Get information about the current session"""

    # Ensure session is initialized
    if not session.initialized:
        await session.initialize(reuse_keys=False)

    return SessionInfo(
        session_id=session.session_id,
        nostr_pubkey=session.nostr_client.get_npub(),
        created_at=session.created_at.isoformat(),
        last_accessed=session.last_accessed.isoformat()
    )



import asyncio
import click
import sys
import logging
from typing import Awaitable, Callable, List

from publsp.blip51.utils import calculate_lease_cost
from publsp.cli.basecli import BaseCLI
from publsp.nostr.client import NostrClient
from publsp.nostr.nip17 import RumorHandler, Nip17Listener
from publsp.marketplace.customer import CustomerHandler, OrderResponseHandler

logger = logging.getLogger(name=__name__)


async def async_prompt(text: str) -> str:
    """Run click.prompt in a thread to avoid blocking the event loop."""
    return await asyncio.to_thread(click.prompt, text)


async def pick_from_list(
    prompt_text: str,
    choices: List[str],
    default_index: int = 0,
) -> str:
    """
    Prompt the user with a list of choices using click.prompt and click.Choice.
    """
    return await asyncio.to_thread(
        click.prompt,
        prompt_text,
        type=click.Choice(choices),
        default=choices[default_index],
    )


class CustomerCLI(BaseCLI):
    def __init__(self, **kwargs):
        # reactor state
        self._running = True

        # core components
        reuse_keys = kwargs.get("reuse_keys")
        self.nostr_client = NostrClient(client_for="customer", reuse_keys=reuse_keys)
        self.rumor_handler = RumorHandler()
        self.nip17_listener = Nip17Listener(
            nostr_client=self.nostr_client,
            rumor_handler=self.rumor_handler,
        )
        self.customer_handler = CustomerHandler(
            nostr_client=self.nostr_client,
            **kwargs,
        )
        self.order_response_handler = OrderResponseHandler(
            customer_handler=self.customer_handler,
            rumor_handler=self.rumor_handler,
            **kwargs,
        )

        # menu → (description, coroutine)
        self.commands: dict[str, tuple[str, Callable[[], Awaitable[None]]]] = {
            "1": ("Show discovered ads", self.cmd_show_ads),
            "2": ("Get liquidity cost breakdown", self.cmd_summarize_costs),
            "3": ("Request a channel", self.cmd_request_channel),
            "4": ("Exit", self.cmd_exit),
        }

    async def startup(self) -> None:
        """Connect relays, fetch ads, and start NIP-17 listener."""
        await self.nostr_client.connect_relays()
        await self.customer_handler.get_ad_info()
        self.nip17_listener.start()
        self.order_response_handler.start()

    async def shutdown(self) -> None:
        """Stop listeners, disconnect, then exit."""
        await self.nip17_listener.stop()
        await self.order_response_handler.stop()
        await self.nostr_client.disconnect_relays()
        # finally terminate the process for real
        sys.exit(0)

    def _render_menu(self) -> None:
        text = "\nChoose an option:\n"
        for key, (desc, _) in self.commands.items():
            text += f"  {key}. {desc}\n"
        click.echo(text)

    # ------------------------------------------
    # Command handlers
    # ------------------------------------------

    async def cmd_show_ads(self) -> None:
        ads = self.customer_handler.active_ads
        if ads and ads.ads:
            click.echo(ads)
        else:
            click.echo("\nNo ads discovered yet.")

    async def cmd_show_ads_short(self) -> None:
        ads = self.customer_handler.active_ads
        if ads and ads.ads:
            click.echo("\nAvailable offers:")
            opts = self.customer_handler.options
            capacity = opts.get('lsp_balance_sat') + opts.get('client_balance_sat')
            click.echo(
                self.customer_handler.summarise_channel_prices(capacity=capacity)
            )
        else:
            click.echo("\nNo ads discovered")

    async def cmd_summarize_costs(self) -> None:
        while True:
            click.echo(
                "\nCost summary:\n"
                "  1. Enter desired capacity (sats)\n"
                "  2. Back\n"
                "  3. Exit\n"
            )
            c = await async_prompt("Choice (1-3)")
            if c == "1":
                cap = await async_prompt("Capacity (sats): ")
                try:
                    summary = self.customer_handler.summarise_channel_prices(
                        capacity=int(cap)
                    )
                    click.echo(str(summary))
                except ValueError:
                    click.echo("Please enter a valid integer.")
            elif c == "2":
                return
            elif c == "3":
                await self.cmd_exit()
                return
            else:
                click.echo("Invalid choice, enter 1-3.")

    async def cmd_request_channel(self) -> None:
        ads = self.customer_handler.active_ads
        if not (ads and ads.ads):
            click.echo("No ads available.")
            return

        ad_ids = list(ads.ads.keys())

        await self.cmd_show_ads_short()

        # replace arrow-list selector with click.Choice
        choice = await pick_from_list(
            "Select Ad ID:",
            ad_ids
        )

        if choice not in ads.ads:
            click.echo(f"Ad ID {choice!r} not found.")
            return

        ad = ads.ads[choice]
        order = self.customer_handler.build_order(ad_id=choice)
        expected_cost = calculate_lease_cost(
            fixed_cost=ad.fixed_cost_sats,
            variable_cost_ppm=ad.variable_cost_ppm,
            capacity=order.total_capacity,
            channel_expiry_blocks=order.channel_expiry_blocks,
            max_channel_expiry_blocks=ad.max_channel_expiry_blocks
        )
        peer_pk = ads.get_nostr_pubkey(ad_id=choice, as_PublicKey=True)
        click.echo(
            f"\nRequesting channel of {order.total_capacity} sats\n"
            f"From {ad.lsp_pubkey} (ad ID: {choice})\n"
            f"Target node: {order.target_pubkey_uri}\n"
            f"Outbound sats: {order.client_balance_sat}, "
            f"Inbound sats: {order.lsp_balance_sat}\n"
            f"Expected order cost: {expected_cost} sats\n"
        )
        ok = await async_prompt("Confirm? [y/n]: ")
        if ok.lower() not in ("y", "yes"):
            click.echo("Cancelled.")
            return
        # register the selected ad
        self.order_response_handler.selected_ad = ad
        # send a dm with the order request for the selected ad
        await self.nostr_client.send_private_msg(
            peer_pk,
            "order request",
            rumor_extra_tags=order.model_dump_tags(),
        )
        click.echo("Order request sent.")

    async def cmd_exit(self) -> None:
        click.echo("Exiting...")
        self._running = False

    # ------------------------------------------
    # Main loop
    # ------------------------------------------

    async def run(self) -> None:
        await self.startup()

        try:
            while self._running:
                self._render_menu()
                choice = await async_prompt("Choice (1-4)")
                handler_entry = self.commands.get(choice)
                if handler_entry:
                    _, handler = handler_entry
                    try:
                        await handler()
                    except Exception as exc:
                        logger.error(f"Command {choice} failed: {exc}")
                else:
                    click.echo("Invalid choice, please enter 1-4")

        except KeyboardInterrupt:
            logger.info("Interrupted, shutting down...")
        finally:
            await self.shutdown()


async def run_customer_cli(**kwargs):
    cli = CustomerCLI(**kwargs)
    await cli.run()



YEARLY_MINED_BLOCKS = int(24*60/10*365)


def calculate_lease_cost(
        fixed_cost: int,
        variable_cost_ppm: int,
        capacity: int,
        channel_expiry_blocks: int,
        max_channel_expiry_blocks: int) -> int:
    """
    the LSP sets a yearly ppm for simplicity, and the customer can request any
    lease time (smaller than the LSPs max) so the yearly ppm on capacity needs
    to be pro-rated to the requested lease time
    """
    variable_cost = variable_cost_ppm * 1e-6 * capacity
    lease_time_factor = channel_expiry_blocks / max_channel_expiry_blocks
    return fixed_cost + round(variable_cost * lease_time_factor)


def calculate_apr(
        fixed_cost: int,
        variable_cost_ppm: int,
        capacity: int,
        max_channel_expiry_blocks: int) -> int:
    """
    assume the channel is closed after the LSP's max lease period and gets
    repurchased at the same price for the number of times the lease
    duration fits in a year
    """
    variable_cost = variable_cost_ppm * 1e-6 * capacity
    num_yearly_renewals = YEARLY_MINED_BLOCKS / max_channel_expiry_blocks
    apr = (fixed_cost + variable_cost) * num_yearly_renewals / capacity * 100
    return round(apr, 2)



from nostr_sdk import (
    Client,
    EventBuilder,
    Kind,
    NostrSigner,
    Tag,
)

from publsp.nostr.keyhandler import KeyHandler
from publsp.nostr.relays import Relays
from publsp.settings import Environment, EnvironmentSettings, NostrSettings

import logging

logger = logging.getLogger(name=__name__)


class NostrClient(Client):
    def __init__(
            self,
            client_for='lsp',
            write_keys: bool = NostrSettings().write_keys,
            reuse_keys: bool = NostrSettings().reuse_keys,
            encrypt_keys: bool = NostrSettings().encrypt_keys):
        self.client_for = client_for
        self.key_handler = KeyHandler(
            client=client_for,
            reuse_keys=reuse_keys,
            write_keys=write_keys,
            encrypt_keys=encrypt_keys)
        self.signer = NostrSigner.keys(self.key_handler.keys)
        super().__init__(self.signer)

    def build_event(self, tags: [Tag], content: str, kind: Kind):
        # build the event with the kind, content, tags and sign with keys
        builder = EventBuilder(kind, content).tags(tags)
        return builder.sign_with_keys(self.key_handler.keys)

    async def connect_relays(self, env: Environment = EnvironmentSettings().environment) -> None:
        # Add relays and connect
        for relay in Relays().get_relays(env=env):
            await self.add_relay(relay)

        await self.connect()

    async def disconnect_relays(self, env: Environment = EnvironmentSettings().environment) -> None:
        for relay in Relays().get_relays(env=env):
            await self.disconnect_relay(relay)

    def get_npub(self) -> str:
        return self.key_handler.keys.public_key().to_bech32()

    def get_public_key_hex(self) -> str:
        return self.key_handler.keys.public_key().to_hex()



import asyncio
import logging

from contextlib import suppress

from nostr_sdk import (
    Event,
    Filter,
    HandleNotification,
    Kind, KindStandard,
    SubscribeOutput,
    Timestamp,
    UnsignedEvent,
    UnwrappedGift,
)
from typing import AsyncIterator, Union

from publsp.blip51.order import Order, OrderResponse, OrderErrorResponse
from publsp.nostr.client import NostrClient
from publsp.ln.requesthandlers import ChannelOpenResponse

logger = logging.getLogger(name=__name__)


class RumorHandler:
    """
    Collects incoming NIP‑17 rumors (gift‑wrapped DMs) into an asyncio.Queue,
    and exposes them via an async‐iterator or async getters.
    """
    def __init__(self) -> None:
        # our buffer of incoming rumors
        self._queue: asyncio.Queue[UnsignedEvent] = asyncio.Queue()

    def on_new_rumor(self, rumor: UnsignedEvent) -> None:
        """
        Called by Nip17NotificationHandler.handle() whenever a new DM arrives.
        """
        # put_nowait so we don't block the NIP‑17 handler
        self._queue.put_nowait(rumor)

    async def get_rumor(self) -> UnsignedEvent:
        """
        Pull one rumor off the queue.
        """
        return await self._queue.get()

    def __aiter__(self) -> AsyncIterator[UnsignedEvent]:
        """
        Allows:
            async for rumor in rumor_handler:
                ...
        """
        return self._rumors()

    async def _rumors(self) -> AsyncIterator[UnsignedEvent]:
        while True:
            rumor = await self._queue.get()
            logger.debug(f'got rumor: {rumor}')
            yield rumor

    async def order_requests(self) -> AsyncIterator[Order]:
        """
        Filters the raw rumors to only yield valid Order requests.
        """
        async for rumor in self:
            tags = rumor.tags().to_vec()
            tag_keys = [t.as_vec()[0] for t in tags]
            order_req_fields = list(Order.model_fields.keys())
            if set(order_req_fields).issubset(tag_keys):
                # build and yield the Order
                order_req = Order.model_from_tags(tags=tags)
                logger.debug(f'rumor is order request: {order_req}')
                yield rumor, order_req

    async def order_responses(self) -> AsyncIterator[Union[OrderResponse, OrderErrorResponse]]:
        """
        Filters the raw rumors to only yield valid Order response.
        """
        async for rumor in self:
            # your existing tag‐based test
            tags = rumor.tags().to_vec()
            tag_keys = [t.as_vec()[0] for t in tags]
            order_resp_fields = list(OrderResponse.model_fields.keys())
            if set(order_resp_fields).issubset(tag_keys):
                # build and yield the Order
                logger.debug('got order response')
                order_resp = OrderResponse.model_from_tags(tags=tags)
                yield rumor, order_resp
            elif set(['error_message', 'code']).issubset(tag_keys):
                logger.debug('got order error response')
                order_err_resp = OrderErrorResponse.model_from_tags(tags=tags)
                yield rumor, order_err_resp

    async def channel_open_responses(self) -> AsyncIterator[Order]:
        """
        Filters the raw rumors to only yield valid channel open responses
        """
        async for rumor in self:
            tags = rumor.tags().to_vec()
            tag_keys = [t.as_vec()[0] for t in tags]
            chan_resp_fields = list(ChannelOpenResponse.model_fields.keys())
            if set(chan_resp_fields).issubset(tag_keys):
                logger.debug('got channel open response')
                chan_open_resp = ChannelOpenResponse.model_from_tags(tags=tags)
                logger.debug(f'rumor is channel open response: {chan_open_resp}')
                yield rumor, chan_open_resp


class Nip17NotificationHandler(HandleNotification):
    def __init__(self, nostr_client: NostrClient, rumor_handler: RumorHandler):
        self.nostr_client = nostr_client
        self.rumor_handler = rumor_handler
        self._ts = Timestamp.now()

    async def handle(self, relay_url, subscription_id, event: Event):
        if event.kind().as_std() == KindStandard.GIFT_WRAP:
            logger.debug("Received a NIP-59 event, decrypting...")
            try:
                # Extract rumor
                unwrapped_gift = await UnwrappedGift\
                    .from_gift_wrap(self.nostr_client.signer, event)
                rumor: UnsignedEvent = unwrapped_gift.rumor()

                # Check timestamp of rumor
                if rumor.created_at().as_secs() >= self._ts.as_secs():
                    rumor_kind = rumor.kind().as_std()
                    if rumor_kind == KindStandard.PRIVATE_DIRECT_MESSAGE:
                        self.rumor_handler.on_new_rumor(rumor)
                    else:
                        logger.debug(f"msg: {rumor.as_json()}")
            except Exception as e:
                logger.error(f"Error during content NIP59 decryption: {e}")

    async def handle_msg(self, relay_url, msg):
        pass


class Nip17Listener:
    """
    Subscribes to GIFT_WRAP DMs for our key,
    unwraps them, and pushes into RumorHandler’s queue.
    """
    def __init__(
            self,
            nostr_client: NostrClient,
            rumor_handler: RumorHandler):
        self.nostr_client = nostr_client
        self.handler = Nip17NotificationHandler(nostr_client, rumor_handler)
        self.filter = (
            Filter()
            .kind(Kind.from_std(KindStandard.GIFT_WRAP))
            .pubkey(nostr_client.key_handler.keys.public_key())
            .limit(0)
        )
        self._sub: Union[SubscribeOutput, None] = None
        self._task: Union[asyncio.Task, None] = None

    async def _run(self):
        # 1) subscribe
        self._sub = await self.nostr_client.subscribe(self.filter)
        # 2) hand off INCOMING events to our notification adapter
        await self.nostr_client.handle_notifications(self.handler)

    def start(self):
        """Begin relaying gift‑wrapped DMs into RumorHandler.queue."""
        if self._task is None or self._task.done():
            self._task = asyncio.create_task(self._run())

    async def stop(self):
        """Tear down the subscription and cancel the background task."""
        if self._task:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

        if self._sub:
            await self.nostr_client.unsubscribe(self._sub.id)
